import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
import matplotlib as mpl
from scipy.interpolate import UnivariateSpline
from IPython.display import clear_output, display, HTML
from scipy import integrate
from ipywidgets.widgets.interaction import interact, interactive

mpl.rcParams['lines.linewidth'] = 2
mpl.rcParams['figure.figsize'] = (10, 6)


def free_response(m=10, c=1, k=100, x0=1, v0=-1, max_time=10):
    """
    returns t, x, v, zeta, omega, omega_d, A
    $\alpha$
    Returns free response of a second order linear ordinary differential equation
    defined by
    :math:`m\ddot{x} + c \dot{x} + k x = 0`
    given initial conditions :math:`x_0` and :math:`\dot{x}_0 = v_0` for
    :math:`0 < t < t_{max}`

    Parameters

    m, c, k:           1) Floats. Mass, damping and stiffness.
    x0, v0:            2) Floats. Initial conditions
    max_time:          3) Float.

    Returns

    t, x, v: 1) Arrays. Time, displacement, and velocity

    :Example:
    >>> free_response()
    (array([  0.00000000e+00,   4.00160064e-03,   8.00320128e-03, ...,
             9.99199680e+00,   9.99599840e+00,   1.00000000e+01]), array([[ 1.        ],
           [ 0.99591926],
           [ 0.9916807 ],
           ...,
           [ 0.56502396],
           [ 0.56123989],
           [ 0.55736747]]), array([[-1.        ],
           [-1.03952678],
           [-1.07887136],
           ...,
           [-0.93454914],
           [-0.95670532],
           [-0.97869947]]), 0.015811388300841896, 3.1622776601683795, 3.1618823507524758, 1.0441611791969838)
"""

    omega = sp.sqrt(k / m)
    zeta = c / 2 / omega / m
    omega_d = omega * sp.sqrt(1 - zeta ** 2)
    A = sp.sqrt(x0 ** 2 + (v0 + omega * zeta * x0) ** 2 / omega_d ** 2)

    #    print('The natural frequency is ', omega, 'rad/s.');
    #    print('The damping ratio is ', zeta);
    #    print('The damped natural frequency is ', omega_d);

    def sdofs_deriv(x_xd, t0, m=m, c=c, k=k):
        x, xd = x_xd
        return [xd, -c / m * xd - k / m * x]

    z0 = np.array([[x0, v0]])
    # Solve for the trajectories
    t = np.linspace(0, max_time, int(250 * max_time))
    z_t = np.asarray([integrate.odeint(sdofs_deriv, z0i, t)
                      for z0i in z0])

    x, y = z_t[:, :].T
    return t, x, y, zeta, omega, omega_d, A


def phase_plot(m=10, c=1, k=100, x0=1, v0=-1, max_time=10):
    '''Phase plot of free response of single degree of freedom system.
    For information on variables see `free_response`'''
    t, x, v, zeta, omega, omega_d, A = free_response(
        m, c, k, x0, v0, max_time)
    fig = plt.figure()
    fig.suptitle('Velocity vs Displacement')
    ax = fig.add_subplot(111)
    ax.set_xlabel('Displacement')
    ax.set_ylabel('Velocity')
    ax.grid('on')
    ax.plot(x, v)


def phase_plot_i(max_time=(1.0, 200.0), v0=(-100, 100, 1.0), m=(1.0, 100.0, 1.0),
                      c=(0.0, 1.0, 0.1), x0=(-100, 100, 1), k=(1.0, 100.0, 1.0)):
    '''Interactive phase plot of free response of single degree of freedom system.
    For information on variables see ``free_response``'''
    w = interactive(phase_plot, max_time=max_time, v0=v0, m=m,
                    c=c, x0=x0, k=k)
    display(w)


def time_plot(m=10, c=1, k=100, x0=1, v0=-1, max_time=100):
    t, x, v, zeta, omega, omega_d, A = free_response(
        m, c, k, x0, v0, max_time)
    fig = plt.figure()
    fig.suptitle('Displacement vs Time')
    ax = fig.add_subplot(111)
    ax.set_xlabel('Time')
    ax.set_ylabel('Displacement')
    ax.grid('on')
    ax.plot(t, x)
    if zeta < 1:
        ax.plot(t, A * sp.exp(-zeta * omega * t), '--', t, -A *
                sp.exp(-zeta * omega * t), '--g', linewidth=1)
        tmin, tmax, xmin, xmax = ax.axis()
        ax.text(.75 * tmax, .95 * (xmax - xmin) + xmin,
                '$\omega$ = %0.2f rad/sec' % (omega))
        ax.text(.75 * tmax, .90 * (xmax - xmin) +
                xmin, '$\zeta$ = %0.2f' % (zeta))
        ax.text(.75 * tmax, .85 * (xmax - xmin) + xmin,
                '$\omega_d$ = %0.2f rad/sec' % (omega_d))
    else:
        tmin, tmax, xmin, xmax = ax.axis()
        ax.text(.75 * tmax, .95 * (xmax - xmin) +
                xmin, '$\zeta$ = %0.2f' % (zeta))
        ax.text(.75 * tmax, .90 * (xmax - xmin) + xmin,
                '$\lambda_1$ = %0.2f' % (zeta * omega - omega * (zeta ** 2 - 1)))
        ax.text(.75 * tmax, .85 * (xmax - xmin) + xmin,
                '$\lambda_2$ = %0.2f' % (zeta * omega + omega * (zeta ** 2 - 1)))


def time_plot_i(max_time=(1.0, 100.0), v0=(-100, 100), m=(1.0, 100.0),
                     c=(0.0, 100.0), x0=(-100, 100), k=(1.0, 100.0)):
    w = interactive(time_plot, max_time=max_time, v0=v0, m=m,
                    c=c, x0=x0, k=k)
    # I'd like to get the sliders to be side by side to take less vertical space
    # cont = widgets.HBox(children = w)
    # print(help(w))
    display(w)


def analytical(m=1, c=0.1, k=1, x0=1, v0=0, n=8, dt=0.05):

    w = np.sqrt(k / m)
    zeta = c / (2 * w * m)  # (1.30)

    wd = w * np.sqrt(1 - zeta**2)  # (1.37)
    t = sp.linspace(0, n * dt, n + 1)

    print('The natural frequency is ', w, 'rad/s.')
    print('The damping ratio is ', zeta)
    print('The damped natural frequency is ', wd)

    if zeta < 1:
        A = np.sqrt(((v0 + zeta * w * x0)**2 + (x0 * wd)**2) / wd**2)  # (1.38)
        phi = np.arctan2(x0 * wd, v0 + zeta * w * x0)  # (1.38)
        x = A * np.exp(-zeta * w * t) * np.sin(wd * t + phi)  # (1.36)
        print('A =', A)
        print('phi =', phi)

    elif zeta == 1:
        a1 = x0  # (1.46)
        a2 = v0 + w * x0  # (1.46)
        print('a1= ', a1)
        print('a2= ', a2)
        x = (a1 + a2 * t) * np.exp(-w * t)  # (1.45)

    else:
        a1 = (-v0 + (-zeta + np.sqrt(zeta**2 - 1)) * w * x0) / \
            (2 * w * np.sqrt(zeta**2 - 1))  # (1.42)
        a2 = (v0 + (zeta + np.sqrt(zeta**2 - 1)) * w * x0) / \
            (2 * w * np.sqrt(zeta**2 - 1))  # (1.43)
        print('a1= ', a1)
        print('a2= ', a2)
        x = np.exp(-zeta * w * t) * (a1 * np.exp(-w * np.sqrt(zeta**2 - 1)
                                                 * t) + a2 * np.exp(w * np.sqrt(zeta**2 - 1) * t))  # (1.41)

    return x


def euler(m=1, c=.1, k=1, x0=1, v0=0, n=8, dt=0.05):
    """
    Returns free response of a second order linear ordinary differential equation
    using the Euler method for integration.

    Parameters
    ----------
    m, c, k: float
        Mass, damping and stiffness.
    x0, v0: float
        Initial conditions
    n: int
        The number of steps
    dt: float
        The step size.

    Returns
    ----------
    t, x, v: array
        Time, displacement, and velocity

    Examples:
    ----------
    >>> euler(m=1, c=.1, k=1, x0=1, v0=0, n=8, dt=0.05)
    (array([ 0.  ,  0.05,  0.1 ,  0.15,  0.2 ,  0.25,  0.3 ,  0.35,  0.4 ]), array([[ 1.        ,  0.        ],
           [ 1.        , -0.05      ],
           [ 0.9975    , -0.09975   ],
           [ 0.9925125 , -0.14912625],
           [ 0.98505619, -0.19800624],
           [ 0.97515588, -0.24626902],
           [ 0.96284242, -0.29379547],
           [ 0.94815265, -0.34046861],
           [ 0.93112922, -0.3861739 ]]))

    """

    # creates the state space matrix
    A = sp.array([[0, 1],
                  [-k / m, -c / m]])
    # creates the x array and set the first line according to the initial
    # conditions
    x = sp.zeros((n + 1, 2))
    x[0] = x0, v0

    for i in range(0, n):
        x[i + 1] = x[i] + dt * A@x[i]

    t = sp.linspace(0, n * dt, n + 1)

    return t, x


def rk4(m=1, c=.1, k=1, x0=1, v0=0, n=8, dt=0.05):
    """
    Returns free response of a second order linear ordinary differential equation
    using the Runge-Kutta method for integration.

    Parameters
    ----------
    m, c, k: float
        Mass, damping and stiffness.
    x0, v0: float
        Initial conditions
    n: int
        The number of steps
    dt: float
        The step size.

    Returns
    ----------
    t, x, v: array
        Time, displacement, and velocity

    Examples:
    ----------
    >>> rk4(m=1, c=.1, k=1, x0=1, v0=0, n=8, dt=0.05)
    (array([ 0.  ,  0.05,  0.1 ,  0.15,  0.2 ,  0.25,  0.3 ,  0.35,  0.4 ]), array([[ 1.        ,  0.        ],
           [ 0.99875234, -0.04985443],
           [ 0.99502078, -0.0993359 ],
           [ 0.98882699, -0.14832292],
           [ 0.98019872, -0.19669582],
           [ 0.96916961, -0.24433704],
           [ 0.95577913, -0.29113145],
           [ 0.94007246, -0.33696662],
           [ 0.92210029, -0.38173305]]))
    """

    t = sp.linspace(0, n * dt, n + 1)
    x = sp.zeros((n + 1, 2))
    x[0, :] = x0, v0
    A = sp.array([[0, 1],
                  [-k / m, -c / m]])

    def f(x_): return A@x_

    for i in range(n):
        k1 = dt * f(x[i])
        k2 = dt * f(x[i] + k1 / 2)
        k3 = dt * f(x[i] + k2 / 2)
        k4 = dt * f(x[i] + k3)
        x[i + 1] = x[i] + (k1 + 2.0 * (k2 + k3) + k4) / 6.0

    return t, x


def euler_beam_modes(n=10, bctype=2, beamparams=sp.array((7.31e10, 1 / 12 * 0.03 * .015 ** 3, 2747, .015 * 0.03, 0.4)),
                     npoints=2001):
    """
    %VTB6_3 Natural frequencies and mass normalized mode shape for an Euler-
    % Bernoulli beam with a chosen boundary condition.
    % [w,x,U]=VTB6_3(n,bctype,bmpar,npoints) will return the nth natural
    % frequency (w) and mode shape (U) of an Euler-Bernoulli beam.
    % If n is a vector, return the coresponding mode shapes and natural
    % frequencies.
    % With no output arguments the modes are ploted.
    % If only one mode is requested, and there are no output arguments, the
    % mode shape is animated.
    % The boundary condition is defined as follows:
    %
    % bctype = 1 free-free
    % bctype = 2 clamped-free
    % bctype = 3 clamped-pinned
    % bctype = 4 clamped-sliding
    % bctype = 5 clamped-clamped
    % bctype = 6 pinned-pinned
    %
    % The beam parameters are input through the vector bmpar:
    % bmpar = [E I rho A L];
    % where the variable names are consistent with Section 6.5 of the
    % text.
    %
    %% Example: 20 cm long aluminum beam with h=1.5 cm, b=3 cm
    %% Animate the 4th mode for free-free boundary conditions
    % E=7.31e10;
    % I=1/12*.03*.015^3;
    % rho=2747;
    % A=.015*.03;
    % L=0.2;
    % vtb6_3(4,1,[E I rho A L]);
    %

    % Copyright Joseph C. Slater, 2007
    % Engineering Vibration Toolbox
    """
    E = beamparams[0]
    I = beamparams[1]
    rho = beamparams[2]
    A = beamparams[3]
    L = beamparams[4]
    if isinstance(n, int):
        ln = n
        n = sp.arange(n) + 1
    else:
        ln = len(n)

    # len=[0:(1/(npoints-1)):1]';  %Normalized length of the beam
    len = sp.linspace(0, 1, npoints)
    x = len * L
    # Determine natural frequencies and mode shapes depending on the
    # boundary condition.
    # Mass simplification. The following was arange_(1,length_(n)).reshape(-1)
    mode_num_range = sp.arange(0, ln)
    Bnl = sp.empty(ln)
    w = sp.empty(ln)
    U = sp.empty([npoints, ln])

    if bctype == 1:
        desc = 'Free-Free '
        Bnllow = sp.array((0, 0, 4.73004074486, 7.8532046241,
                           10.995607838, 14.1371654913, 17.2787596574))
        for i in mode_num_range:
            if n[i] > 7:
                Bnl[i] = (2 * n[i] - 3) * sp.pi / 2
            else:
                Bnl[i] = Bnllow[i]
        for i in mode_num_range:
            if n[i] == 1:
                w[i] = 0
                U[:, i] = 1 + len * 0
            elif n[i] == 2:
                w[i] = 0
                U[:, i] = len - 0.5
            else:
                sig = (sp.cosh(Bnl[i]) - sp.cos(Bnl[i])) / \
                      (sp.sinh(Bnl[i]) - sp.sin(Bnl[i]))
                w[i] = (Bnl[i] ** 2) * sp.sqrt(E * I / (rho * A * L ** 4))
                b = Bnl[i] * len
                U[:, i] = sp.cosh(b) + sp.cos(b) - sig * \
                    (sp.sinh(b) + sp.sin(b))
    elif bctype == 2:
        desc = 'Clamped-Free '
        Bnllow = sp.array((1.88, 4.69, 7.85, 10.99, 14.14))
        for i in mode_num_range:
            if n[i] > 4:
                Bnl[i] = (2 * n[i] - 1) * sp.pi / 2
            else:
                Bnl[i] = Bnllow[i]

        for i in mode_num_range:
            sig = (sp.sinh(Bnl[i]) - sp.sin(Bnl[i])) / \
                  (sp.cosh(Bnl[i]) - sp.cos(Bnl[i]))
            w[i] = (Bnl[i] ** 2) * sp.sqrt(E * I / (rho * A * L ** 4))
            b = Bnl[i] * len
            # plt.plot(x,(sp.cosh(b) - sp.cos(b) - sig * (sp.sinh(b) - sp.sin(b))))
            U[:, i] = sp.cosh(b) - sp.cos(b) - sig * (sp.sinh(b) - sp.sin(b))

    elif bctype == 3:
        desc = 'Clamped-Pinned '
        Bnllow = sp.array((3.93, 7.07, 10.21, 13.35, 16.49))
        for i in mode_num_range:
            if n[i] > 4:
                Bnl[i] = (4 * n[i] + 1) * sp.pi / 4
            else:
                Bnl[i] = Bnllow[i]
        for i in mode_num_range:
            sig = (sp.cosh(Bnl[i]) - sp.cos(Bnl[i])) / \
                  (sp.sinh(Bnl[i]) - sp.sin(Bnl[i]))
            w[i] = (Bnl[i] ** 2) * sp.sqrt(E * I / (rho * A * L ** 4))
            b = Bnl[i] * len
            U[:, i] = sp.cosh(b) - sp.cos(b) - sig * (sp.sinh(b) - sp.sin(b))
    elif bctype == 4:
        desc = 'Clamped-Sliding '
        Bnllow = sp.array((2.37, 5.5, 8.64, 11.78, 14.92))
        for i in mode_num_range:
            if n[i] > 4:
                Bnl[i] = (4 * n[i] - 1) * sp.pi / 4
            else:
                Bnl[i] = Bnllow[i]
        for i in mode_num_range:
            sig = (sp.sinh(Bnl[i]) + sp.sin(Bnl[i])) / \
                  (sp.cosh(Bnl[i]) - sp.cos(Bnl[i]))
            w[i] = (Bnl[i] ** 2) * sp.sqrt(E * I / (rho * A * L ** 4))
            b = Bnl[i] * len
            U[:, i] = sp.cosh(b) - sp.cos(b) - sig * (sp.sinh(b) - sp.sin(b))
    elif bctype == 5:
        desc = 'Clamped-Clamped '
        Bnllow = sp.array((4.73, 7.85, 11, 14.14, 17.28))
        for i in mode_num_range:
            if n[i] > 4:
                Bnl[i] = (2 * n[i] + 1) * sp.pi / 2
            else:
                Bnl[i] = Bnllow[i]
        for i in mode_num_range:
            sig = (sp.cosh(Bnl[i]) - sp.cos(Bnl[i])) / \
                  (sp.sinh(Bnl[i]) - sp.sin(Bnl[i]))
            w[i] = (Bnl[i] ** 2) * sp.sqrt(E * I / (rho * A * L ** 4))
            b = Bnl[i] * len
            U[:, i] = sp.cosh(b) - sp.cos(b) - sig * (sp.sinh(b) - sp.sin(b))
    elif bctype == 6:
        desc = 'Pinned-Pinned '
        for i in mode_num_range:
            Bnl[i] = n[i] * sp.pi
            w[i] = (Bnl[i] ** 2) * sp.sqrt(E * I / (rho * A * L ** 4))
            U[:, i] = sp.sin(Bnl[i] * len)

    # Mass Normalization of mode shapes
    for i in mode_num_range:
        U[:, i] = U[:, i] / sp.sqrt(sp.dot(U[:, i], U[:, i]) * rho * A * L)

    """
    ppause=0
    x=len * L
    if nargout == 0:
        if length_(n) != 1:
            for i in arange_(1,length_(n)).reshape(-1):
                plot_(x,U[:,i])
                axis_([0,L,min_(min_(U)),max_(max_(U))])
                figure_(gcf)
                title_([desc,char('  '),char('Mode '),int2str_(i),char('     Natural Frequency = '),num2str_(w[i]),char(' rad/s')])
                ylabel_(char('Modal Amplitude'))
                xlabel_(char('Length along bar - x'))
                grid_(char('on'))
                disp_(char('Press return to continue'))
                pause
        else:
            nsteps=50
            clf
            step=2 * pi / (nsteps)
            i=arange_(0,(2 * pi - step),step)
            hold_(char('off'))
            handle=uicontrol_(char('style'),char('pushbutton'),char('units'),char('normal'),char('backgroundcolor'),char('red'),char('position'),[0.94,0.94,0.05,0.05],char('String'),char('Stop'),char('callback'),char('global stopstop;stopstop=1;'))
            handle2=uicontrol_(char('style'),char('pushbutton'),char('units'),char('normal'),char('backgroundcolor'),char('yellow'),char('position'),[0.94,0.87,0.05,0.05],char('String'),char('Pause'),char('callback'),char('global ppause;ppause=1;'))
            handle3=uicontrol_(char('style'),char('pushbutton'),char('units'),char('normal'),char('backgroundcolor'),char('green'),char('position'),[0.94,0.8,0.05,0.05],char('String'),char('Resume'),char('callback'),char('global ppause;ppause=0;'))
            stopstop=0
            bb=0
            while stopstop == 0 and bb < 100:

                bb=bb + 1
                for ii in [i].reshape(-1):
                    while ppause == 1:

                        pause_(0.01)
                        if stopstop == 1:
                            delete_(handle)
                            delete_(handle2)
                            delete_(handle3)
                            return w,x,U

                    plot_(x,U[:,1] * sp.cos(ii))
                    axis_([0,L,- max_(abs_(U)),max_(abs_(U))])
                    grid_(char('on'))
                    figure_(gcf)
                    title_([desc,char('  '),char('Mode '),int2str_(n),char('     \\omega_n = '),num2str_(w[1]),char(' rad/s')])
                    ylabel_(char('Modal Amplitude'))
                    xlabel_(char('Length along bar - x'))
                    drawnow

            clear_(char('stopstop'))
            delete_(handle)
            delete_(handle2)
            delete_(handle3)
    """
    return w, x, U


def euler_beam_frf(xin=0.22, xout=0.22, fmin=0.0, fmax=1000.0, zeta=0.02,
                   beamparams=sp.array((7.31e10, 1 / 12 * 0.03 * .015 ** 3, 2747, .015 * 0.03, 0.4)), bctype=2):
    print(fmin)
    E = beamparams[0]
    I = beamparams[1]
    rho = beamparams[2]
    A = beamparams[3]
    L = beamparams[4]
    np = 2001
    i = 0
    w = sp.linspace(fmin, fmax, 2001) * 2 * sp.pi
    if min([xin, xout]) < 0 or max([xin, xout]) > L:
        disp_(char('One or both locations are not on the beam'))
        return
    wn = sp.array((0, 0))
    # The number 100 is arbitrarily large and unjustified.
    a = sp.empty([np, 100], dtype=complex)
    f = sp.empty(100)

    while wn[-1] < 1.3 * (fmax * 2 * sp.pi):
        i = i + 1
        # legtext[i + 1]=[char('Contribution of mode '),num2str_(i)]
        wn, xx, U = euler_beam_modes(i, bctype, beamparams, 5000)
        spl = UnivariateSpline(xx, U[:, i - 1])
        Uin = spl(xin)
        Uout = spl(xout)
        # Uin=spline_(xx,U,xin)
        # Uout=spline_(xx,U,xout)

        # print(wn[-1])
        # print(w)
        a[:, i - 1] = rho * A * Uin * Uout / \
            (wn[-1] ** 2 - w ** 2 + 2 * zeta * wn[-1] * w * sp.sqrt(-1))
        # print(a[0:10,i])
        # plt.plot(sp.log10(sp.absolute(a[:,i])))
        # input("Press Enter to continue...")
        f[i] = wn[-1] / 2 / sp.pi
    a = a[:, 0:i]
    plt.subplot(211)
    plt.plot(w / 2 / sp.pi, 20 * sp.log10(sp.absolute(sp.sum(a, axis=1))), '-')
    plt.hold('on')
    plt.plot(w / 2 / sp.pi, 20 * sp.log10(sp.absolute(a)), '-')
    plt.grid('on')
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('FRF (dB)')
    axlim = plt.axis()

    plt.axis(
        axlim + sp.array([0, 0, -0.1 * (axlim[3] - axlim[2]), 0.1 * (axlim[3] - axlim[2])]))

    plt.subplot(212)
    plt.plot(w / 2 / sp.pi, sp.unwrap(sp.angle(sp.sum(a, axis=1))) /
             sp.pi * 180, '-')
    plt.hold('on')
    plt.plot(w / 2 / sp.pi, sp.unwrap(sp.angle(a)) / sp.pi * 180, '-')
    plt.grid('on')
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (deg)')
    axlim = plt.axis()
    plt.axis(
        axlim + sp.array([0, 0, -0.1 * (axlim[3] - axlim[2]), 0.1 * (axlim[3] - axlim[2])]))

    fout = w / 2 / sp.pi
    H = a
    return fout, H


def ebf(xin, xout, fmin, fmax, zeta):
    _, _ = euler_beam_frf(xin, xout, fmin, fmax, zeta)
    return


def ebf1(xin, xout):
    _, _ = euler_beam_frf(xin, xout)
    return


# def
# euler_beam_frf(xin=0.22,xout=0.22,fmin=0.0,fmax=1000.0,beamparams=sp.array((7.31e10,
# 1/12*0.03*.015**3, 2747, .015*0.03, 0.4)),


def frfplot(f, H):
    plt.subplot(211)
    plt.plot(f, 20 * sp.log10(sp.absolute(sp.sum(H, axis=1))), '-')
    plt.grid('on')
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('FRF (dB)')
    axlim = plt.axis()
    plt.axis(
        axlim + sp.array([0, 0, -0.1 * (axlim[3] - axlim[2]), 0.1 * (axlim[3] - axlim[2])]))

    plt.subplot(212)
    plt.plot(f, sp.unwrap(sp.angle(sp.sum(H, axis=1))) / sp.pi * 180, '-')
    plt.grid('on')
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (deg)')
    axlim = plt.axis()
    plt.axis(
        axlim + sp.array([0, 0, -0.1 * (axlim[3] - axlim[2]), 0.1 * (axlim[3] - axlim[2])]))


def response(xdd, f, t, x0, v0):
    """
    returns t, x, v
    :math:`\ddot{x} = g(x,v) + f(t)`
    given initial conditions :math:`x_0` and :math:`\dot{x}_0 = v_0` for the time `t`

*** Function hasn't been written yet... work in progress. Mimics vtb1_3
    Parameters

    m, c, k:           1) Floats. Mass, damping and stiffness.
    x0, v0:            2) Floats. Initial conditions
    max_time:          3) Float. end time or response to be returned

    Returns

    t, x, v: 1) Arrays. Time, displacement, and velocity

    :Example:
    >>> free_response()
    (array([  0.00000000e+00,   4.00160064e-03,   8.00320128e-03, ...,
             9.99199680e+00,   9.99599840e+00,   1.00000000e+01]), array([[ 1.        ],
           [ 0.99591926],
           [ 0.9916807 ],
           ...,
           [ 0.56502396],
           [ 0.56123989],
           [ 0.55736747]]), array([[-1.        ],
           [-1.03952678],
           [-1.07887136],
           ...,
           [-0.93454914],
           [-0.95670532],
           [-0.97869947]]), 0.015811388300841896, 3.1622776601683795, 3.1618823507524758, 1.0441611791969838)
    """

    omega = sp.sqrt(k / m)
    zeta = c / 2 / omega / m
    omega_d = omega * sp.sqrt(1 - zeta**2)
    A = sp.sqrt(x0**2 + (v0 + omega * zeta * x0)**2 / omega_d**2)
#    print('The natural frequency is ', omega, 'rad/s.');
#    print('The damping ratio is ', zeta);
#    print('The damped natural frequency is ', omega_d);

    def sdofs_deriv(x_xd, t0, m=m, c=c, k=k):
        x, xd = x_xd
        return [xd, -c / m * xd - k / m * x]

    z0 = np.array([[x0, v0]])
    # Solve for the trajectories
    t = np.linspace(0, max_time, int(250 * max_time))
    z_t = np.asarray([integrate.odeint(sdofs_deriv, z0i, t)
                      for z0i in z0])

    x, y = z_t[:, :].T
    return t, x, y, zeta, omega, omega_d, A


def forced_analytical(m=10, k=100, x0=1, v0=0,
                           wdr=0.5, F0=10, tf=100):

    """
    Returns the response of an undamped single degree of freedom system
    to a sinusoidal input with amplitude F0 and frequency wdr.

    Parameters
    ----------
    m, k: float
        Mass and stiffness
    x0, v0: float
        Initial conditions
    wdr:
        Force frequency
    F0: float
        Force magnitude
    tf: float
        End time

    Returns
    ----------
    t, x: array
        Time and displacement

    Examples:
    >>> forced_analytical(m=10, k=100, x0=1, v0=0, wdr=0.5, F0=10, tf=100)
    (array([  0.00000000e+00,   1.25000156e-04,   2.50000313e-04, ...,
             9.99997500e+01,   9.99998750e+01,   1.00000000e+02]), array([ 1.        ,  0.99999993,  0.99999972, ..., -0.32885349,
           -0.32916361, -0.32947367]))"""

    t = sp.linspace(0, tf, tf/0.000125)

    f0 = F0 / m
    w = sp.sqrt(k / m)
    x = (v0 / w * sp.sin(w * t) +
         (x0 - f0 / (w**2 - wdr**2)) * sp.cos(w * t) +
         f0 / (w**2 - wdr**2) * sp.cos(wdr * t))   # (2.11)

    return t, x


def forced_response(m=10, c=0, k=100, x0=1, v0=0,
                        wdr=0.5, F0=10, max_time=100):
    """
    Returns the the response of an underdamped single degree of
    freedom system to a sinusoidal input with amplitude F0 and
    frequency wdr.

    Parameters
    ----------
    m, c, k: float
        Mass and stiffness
    x0, v0: float
        Initial conditions
    wdr:
        Force frequency
    F0: float
        Force magnitude
    max_time: float
        End time

    Returns
    ----------
    t, x, y: array
        Time, displacement and velocity

    Examples:
    >>> f = forced_response(m=10, c=0, k=100, x0=1, v0=0, wdr=0.5, F0=10, max_time=100)
    >>> f[0][0]
    0.0"""

    def sdofs_deriv(x_xd, t, m=m, c=c, k=k):
        x, xd = x_xd
        return [xd, (F0*sp.cos(wdr*t) / m) - (c / m) * xd - (k / m) * x]

    z0 = sp.array([x0, v0])
    # Solve for the trajectories
    t = sp.linspace(0, max_time, int(250 * max_time))
    z_t = integrate.odeint(sdofs_deriv, z0, t)

    x, y = z_t.T
    return t, x, y


def steady_state_response(zs, rmin, rmax):
    """
    Returns a plot with the steady state response of a
    single degree of freedom damped system.

    Parameters
    ----------
    zs: array
        Array with the damping values
    rmin, rmax: float
        Minimum and maximum frequency ratio

    Returns
    ----------
    r: Array
        Array containing the values for the frequency ratio
    A: Array
        Array containing the values for anmplitude

        Plot with steady state magnitude and phase

    Examples:
    >>> r, A = steady_state_response([0.1, 0.3, 0.8], 0, 2)
    >>> A[10]
    (0.98423159842039087-0.15988334018879749j)
    """

    if not isinstance(zs, list):
        zs = [zs]
    r = sp.linspace(rmin, rmax, 100*(rmax-rmin))
    A0 = sp.zeros((len(zs), len(r)), complex)
    for z in enumerate(zs):
        A0[z[0]] = (1/(1 - r**2 + 2*1j*r*z[1]))

    fig = plt.figure()
    ax1 = fig.add_subplot(211)
    ax2 = fig.add_subplot(212, sharex=ax1)
    plt.tight_layout()

    ax1.set_ylabel('Normalized Amplitude (dB)')
    ax1.set_title('Normalized Amplitude vs Frequency Ratio')

    ax2.set_xlabel('Frequency Ratio')
    ax2.set_ylabel('Phase Lag (deg)')
    ax2.set_title('Phase vs Frequency Ratio')

    for A in A0:
        ax1.plot(r, (sp.absolute(A)))
        ax2.plot(r, -sp.angle(A)/sp.pi*180)

    ax1.legend((['$\zeta$ = ' + (str(s)) for s in zs]))

    _ = plt.show()

    return r, A


def transmissibility(zs, rmin, rmax):
    """
    Returns a plot Displacement transmissibility ratio
    and force transmissibility ratio of a single degree
    of freedom damped system.

    Parameters
    ----------
    zs: array
        Array with the damping values
    rmin, rmax: float
        Minimum and maximum frequency ratio

    Returns
    ----------
    r: Array
        Array containing the values for the frequency ratio
    D: Array
        Array containing the values for displacement
    F: Array
        Array containing the values for force

        Plot with Displacement transmissibility ratio
        and force transmissibility ratio

    Examples:
    >>> r, D, F = transmissibility([0.01, 0.05, 0.1, 0.25, 0.5, 0.7], 0, 2)
    >>> D[10]
    1.0100027508815634
    """

    if not isinstance(zs, list):
        zs = [zs]
    r = sp.linspace(rmin, rmax, 100*(rmax-rmin))
    DT = sp.zeros((len(zs), len(r)))
    for z in enumerate(zs):
        # 2.71
        DT[z[0]] = ((1 + (2 * z[1] * r)**2) /
                    ((1 - r**2)**2 + (2 * z[1] * r)**2))**0.5

    FT = (r**2)*DT

    fig = plt.figure()
    ax1 = fig.add_subplot(211)
    ax2 = fig.add_subplot(212, sharex=ax1)
    plt.tight_layout()

    ax1.set_ylabel('Displacement Transmissibility Ratio (dB)')
    ax1.set_title('Displacement Transmissibility Ratio vs Frequency Ratio (X/Y)')
    ax1.set_ylim(0, 6)

    ax2.set_xlabel('Frequency Ratio')
    ax2.set_ylabel('Force Transmissibility Ratio (dB)')
    ax2.set_title('Force Transmissibility Ratio versus Frequency Ratio (F_T/kY)')
    ax2.set_yscale("log")
    ax2.set_ylim(0.01, 50)

    for D in DT:
        ax1.plot(r, D)
    for F in FT:
        ax2.plot(r, F)

    ax1.legend((['$\zeta$ = ' + (str(s)) for s in zs]))

    _ = plt.show()

    return r, D, F


def rotating_unbalance(m, m0, e, zs, rmin, rmax, normalized=True):
    """
    Returns a plot Displacement of a system with rotating
    unbalance.

    Parameters
    ----------
    m: float
        Mass of the system
    m0, e: float
        Mass and eccentricity of the unbalance.
    zs: array
        Array with the damping values
    rmin, rmax: float
        Minimum and maximum frequency ratio
    normalized: bool
        If true, the displacement is normalized (m*X/(m0*e))

    Returns
    ----------
    r: Array
        Array containing the values for the frequency ratio
    Xn: Array
        Array containing the values for displacement

        Plot with Displacement displacement and phase
        for a system with rotating unbalance.

    Examples:
    >>> r, Xn = rotating_unbalance(m=1, m0=0.5, e=0.1, zs=[0.1, 0.25, 0.707, 1], rmin=0, rmax=3.5, normalized=True)
    >>> Xn[1][10]
    (0.10104614704226758-0.0051182602098315527j)
    """

    if not isinstance(zs, list):
        zs = [zs]
    r = sp.linspace(rmin, rmax, 100*(rmax-rmin))
    Xn = sp.zeros((len(zs), len(r)), complex)
    for z in enumerate(zs):
        Xn[z[0]] = (r / (1 - r**2 + 2*1j*r*z[1]))

    fig = plt.figure()
    ax1 = fig.add_subplot(211)
    ax2 = fig.add_subplot(212, sharex=ax1)
    plt.tight_layout()

    if normalized==False:
        Xn = Xn * (m0 * e / m)
        ax1.set_ylabel('Displacement Magnitude')
        ax1.set_title('Displacement Magnitude vs Frequency Ratio')
    else:
        ax1.set_ylabel('Normalized Displacement Magnitude')
        ax1.set_title('Normalized Displacement Magnitude vs Frequency Ratio')

    ax2.set_xlabel('Frequency Ratio')
    ax2.set_ylabel('Phase')
    ax2.set_title('Phase vs Frequency Ratio')

    for X_z in Xn:
        ax1.plot(r, sp.absolute(X_z))
        ax2.plot(r, -sp.angle(X_z)/sp.pi*180)

    ax1.legend((['$\zeta$ = ' + (str(s)) for s in zs]))

    _ = plt.show()

    return r, Xn


def impulse_response(m, c, k, Fo, max_time):
    """
    Returns a plot with the response of the system to an
    impulse of magnitude Fo (N.s).

    Parameters
    ----------
    m, c, k: float
        Mass, damping and stiffness.
    Fo: float
        Force applied over time (units N.s)
    max_time: float
        End time

    Returns
    ----------
    t: Array
        Array containing the values for the time
    x: Array
        Array containing the values for displacement

        Plot with the response of the system to an
        impulse of magnitude Fo (N.s).

    Examples:
    >>> t, x = impulse_response(m=100, c=20, k=2000, Fo=10, max_time=100)
    >>> x[10]
    0.0039629845398805623
    """

    t = sp.linspace(0, max_time, int(250 * max_time))

    wn = sp.sqrt(k / m)
    zeta = c / (2 * wn * m)
    wd = wn * sp.sqrt(1 - zeta**2)
    fo = Fo / m

    x = fo / (wd * sp.exp(zeta * wn * t)) * sp.sin(wd * t)

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.set_xlabel('Time')
    ax1.set_ylabel('Displacement')
    ax1.set_title('Displacement vs Time')
    ax1.plot(t, x)

    _ = plt.show()

    return t, x


def step_response(m, c, k, Fo, max_time):
    """
    Returns a plot with the response of the system to an
    step of magnitude Fo.

    Parameters
    ----------
    m, c, k: float
        Mass, damping and stiffness.
    Fo: float
        Force applied
    max_time: float
        End time

    Returns
    ----------
    t: Array
        Array containing the values for the time
    x: Array
        Array containing the values for displacement

        Plot with the response of the system to an
        step of magnitude Fo.

    Examples:
    >>> t, x = step_response(m=100, c=20, k=2000, Fo=10, max_time=100)
    >>> x[10]
    7.9581008173000833e-05
    """

    t = sp.linspace(0, max_time, int(250 * max_time))

    wn = sp.sqrt(k / m)
    zeta = c / (2 * wn * m)
    wd = wn * sp.sqrt(1 - zeta**2)
    fo = Fo / m

    if zeta != 1:
        phi = sp.arctan(zeta / sp.sqrt(1 - zeta**2))

    if 0 < zeta < 1:
        x = fo / wn**2 * (1 - wn / wd*sp.exp(-zeta * wn * t)*sp.cos(wd * t - phi))
    elif zeta == 1:
        lam = -wn
        A1 = -fo / wn**2
        A2 = -A1 * lam
        x = fo / wn**2 + A1 * sp.exp(lam * t) + A2 * t * sp.exp(lam * t)
    elif zeta > 1:
        lam1 = -zeta * wn - wn * sp.sqrt(zeta**2 - 1)
        lam2 = -zeta * wn + wn * sp.sqrt(zeta**2 - 1)
        A2 = fo / (wn**2 * (lam2 / lam1 - 1))
        A1 = -lam2 / lam1 * A2
        x = fo / wn**2 + A1 * sp.exp(lam1 * t) + A2 * sp.exp(lam2 * t)
    else:
        raise ValueError('Zeta should be greater than zero')

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.set_xlabel('Time')
    ax1.set_ylabel('Displacement')
    ax1.set_title('Displacement vs Time')
    ax1.plot(t, x)

    _ = plt.show()

    return t, x


def fourier_series(dat, t, n):
    """
    Fourier series approximation to a function.
    returns Fourier coefficients of a function.
    The coefficients are numerical approximations of the true
    coefficients.

    Parameters
    ----------
    dat: array
        Array of data representing the function.
    t: array
        Corresponding time array.
    n: int
        The desired number of terms to use in the
        Fourier series.

    Returns
    ----------
    a, b: tuple
        Tuple containing arrays with the Fourier coefficients.
        The function also produces a plot of the approximation.

    Examples:
    >>> f = sp.hstack((sp.arange(-1, 1, .04), sp.arange(1, -1, -.04)))
    >>> f += 1
    >>> t = sp.arange(0, len(f))/len(f)
    >>> a, b = fourier_series(f, t, 5)
    >>> a[0]
    2.0
    """

    len_ = len(dat)/2
    fs = (sp.fft(dat))/len_
    a0 = fs[0]
    a = sp.real(sp.hstack((a0, fs[1:len(fs/2)])))
    b = -sp.imag(fs[1:len(fs/2)])
    len_ *= 2
    dt = 2*sp.pi/len_
    tp = sp.arange(0, 2*sp.pi, dt)
    dataapprox = a[0]/2 + sp.zeros_like(dat)
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.plot(t, dat)

    for i in range(1, n):
        newdat = a[i]*sp.cos(tp*i) + b[i]*sp.sin(tp*i)
        dataapprox += newdat
        if i == n-1:
            ax1.plot(t, newdat)

    ax1.plot(t, dataapprox)

    _ = plt.show()

    return a, b


def response_spectrum(f):
    """
    Will display the response spectrum to a partial ramp
    input(see Figure 3.13) for the system with natural frequency
    f (in Hz) and no damping.

    Parameters
    ----------
    f: float
        Natural frequency.

    Returns
    ----------
    t, rs: tuple
        Tuple with time and response arrays. It also returns
        a plot with the response spectrum.
    Examples:
    >>> t, rs = response_spectrum(10)
    >>> rs[10]
    1.6285602401720802
    """

    t = sp.linspace(.001 * 4 / f, 10 / f, 200)
    w = 2 * sp.pi * f

    one = sp.ones_like(t)

    rs1 = one / (w * t)
    rs2 = sp.sqrt(2 * (1 - sp.cos(w * t)))

    rs = one + rs1 * rs2

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.set_xlabel('Rise time (t_1)')
    ax1.set_ylabel('Dimensionless maximum response - (xk/Fo)max')
    ax1.set_title('Response spectrum of a SDOF system with f = %s Hz' % f)
    ax1.plot(t, rs)

    _ = plt.show()

    return t, rs


def fourier_approximation(a0, aodd, aeven, bodd, beven, N, T):
    """
    Plot the Fourier series defined by:
    N is the number of terms.

    Parameters
    ----------
    a0: float or str
        a0 Fourier coefficient.
    aodd: float or str
        an Fourier coefficient for n odd.
    aeven: float or str
        an Fourier coefficient for n even.
    bodd: float or str
        bn Fourier coefficient for n odd
    beven: float or str
        bn Fourier coefficient for n even

    Returns
    ----------
    t, F: tuple
        Tuple with time and F(t). It also returns
        a plot with the Fourier approximation.
    Examples:
    >>> # Square wave
    >>> t, F = fourier_approximation(-1, 0, 0, '-3*(-1+(-1)**n)/n/pi', '-3*(-1+(-1)**n)/n/pi', 20, 2)
    >>> F[10]
    1.2697210294282535
    >>> # Triangular wave
    >>> t, F = fourier_approximation(0,'-8/pi**2/n**2',0,0,0,20,10)
    >>> F[10]
    -0.90234928911935097
    """
    args = [str(arg) for arg in [a0, aodd, aeven, bodd, beven]] # change to str
    a0, aodd, aeven, bodd, beven = args

    dt = min(T/400, T/10*N)
    t = sp.arange(0, T*3, dt)
    F = 0*t + eval(a0)/2
    pi = sp.pi

    for n in range(1, N):
        if n % 2 == 0:
            a = aeven
            b = beven
        else:
            a = aodd
            b = bodd
        F = F + eval(a)*sp.cos(n*2*sp.pi*t/T) + eval(b)*sp.sin(n*2*sp.pi*t/T)

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.set_xlabel('Time, t')
    ax1.set_ylabel('F(t)')
    ax1.plot(t, F)

    _ = plt.show()

    return t, F


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    #import vibration-toolbox as vtb


    # doctest.run_docstring_examples(frfest,globals(),optionflags=doctest.ELLIPSIS)
    # doctest.run_docstring_examples(asd,globals(),optionflags=doctest.ELLIPSIS)
    """ What this does.
    python (name of this file)  -v
    will test all of the examples in the help.

    Leaving off -v will run the tests without any output. Success will return nothing.

    See the doctest section of the python manual.
    https://docs.python.org/3.5/library/doctest.html
    """



import scipy as sp
import scipy.linalg as la
import scipy.signal as signal
import matplotlib as mpl

mpl.rcParams['lines.linewidth'] = 2
mpl.rcParams['figure.figsize'] = (10, 6)


def eigen(A, B=None):
    """
    This function is used to sort eigenvalues and eigenvectors
    e.g. for a given system linalg.eig will return eingenvalues as:
    (array([ 0. +89.4j,  0. -89.4j,  0. +89.4j,  0. -89.4j,  0.+983.2j,  0.-983.2j,  0. +40.7j,  0. -40.7j])
    This function will sort this eigenvalues as:
    (array([ 0. +40.7j,  0. +89.4j,  0. +89.4j,  0.+983.2j,  0. -40.7j,  0. -89.4j,  0. -89.4j,  0.-983.2j])
    Correspondent eigenvectors will follow the same order.

    Parameters
    ----------
    A: array
        A complex or real matrix whose eigenvalues and eigenvectors
        will be computed.
    B: float or str
        Right-hand side matrix in a generalized eigenvalue problem.
        Default is None, identity matrix is assumed.

    Returns
    ----------
    evalues: array
        Sorted eigenvalues
    evectors: array
        Sorted eigenvalues

    Examples:
    >>> L = sp.array([[2, -1, 0],
    ...               [-4, 8, -4],
    ...               [0, -4, 4]])
    >>> lam, P = eigen(L)
    >>> lam
    array([  0.56258062+0.j,   2.63206172+0.j,  10.80535766+0.j])
    """
    if B is None:
        evalues, evectors = la.eig(A)
    else:
        evalues, evectors = la.eig(A, B)

    if all(eigs == 0 for eigs in evalues.imag):
        if all(eigs > 0 for eigs in evalues.real):
            idxp = evalues.real.argsort()  # positive in increasing order
            idxn = sp.array([], dtype=int)
        else:
            idxp = evalues.real.argsort()[int(len(evalues)/2):]  # positive in increasing order
            idxn = evalues.real.argsort()[int(len(evalues)/2) - 1::-1]  # negative in decreasing order

    else:
        idxp = evalues.imag.argsort()[int(len(evalues)/2):]  # positive in increasing order
        idxn = evalues.imag.argsort()[int(len(evalues)/2) - 1::-1]  # negative in decreasing order

    idx = sp.hstack([idxp, idxn])

    return evalues[idx], evectors[:, idx]


def normalize(X, Y):
    """
    This function is used to normalize vectors of the matrix
    Y with respect to X so that Y.T @ X = I (identity).
    This is used to normalize the matrix with the left eigenvectors.

    Parameters
    ----------
    X: array
        A complex or real matrix
    Y: array
        A complex or real matrix to be normalized

    Returns
    ----------
    Yn: array
        Normalized matrix

    Examples:
    >>> X = sp.array([[ 0.84+0.j  ,  0.14-0.j  ,  0.84-0.j  ,  0.14+0.j  ],
    ...               [ 0.01-0.3j ,  0.00+0.15j,  0.01+0.3j ,  0.00-0.15j],
    ...               [-0.09+0.42j, -0.01+0.65j, -0.09-0.42j, -0.01-0.65j],
    ...               [ 0.15+0.04j, -0.74+0.j  ,  0.15-0.04j, -0.74-0.j  ]])
    >>> Y = sp.array([[-0.03-0.41j,  0.04+0.1j , -0.03+0.41j,  0.04-0.1j ],
    ...            [ 0.88+0.j  ,  0.68+0.j  ,  0.88-0.j  ,  0.68-0.j  ],
    ...            [-0.21-0.j  ,  0.47+0.05j, -0.21+0.j  ,  0.47-0.05j],
    ...            [ 0.00-0.08j,  0.05-0.54j,  0.00+0.08j,  0.05+0.54j]])
    >>> Yn = normalize(X, Y)
    >>> Yn
    array([[ 0.57822773 -4.69882840e-02j,  0.11696967 -5.85231773e-02j,
             0.57822773 +4.69882840e-02j,  0.11696967 +5.85231773e-02j],
           [ 0.00998912 +1.24180506e+00j, -0.06879320 -8.22911025e-01j,
             0.00998912 -1.24180506e+00j, -0.06879320 +8.22911025e-01j],
           [-0.00238377 -2.96339843e-01j,  0.01295993 -5.73835061e-01j,
            -0.00238377 +2.96339843e-01j,  0.01295993 +5.73835061e-01j],
           [ 0.11289137 -9.08101611e-04j, -0.65854649 -5.87827297e-03j,
             0.11289137 +9.08101611e-04j, -0.65854649 +5.87827297e-03j]])
    """

    Yn = sp.zeros_like(X)
    YTX = Y.T @ X # normalize y so that Y.T @ X will return I
    factors = [1/a for a in sp.diag(YTX)]
    # multiply each column in y by a factor in 'factors'
    for col in enumerate(Y.T):
        Yn[col[0]] = col[1]*factors[col[0]]
    Yn = Yn.T

    return Yn


def modes_system_undamped(M, K):
    """
    This function will return the natural frequencies (w),
    eigenvectors (P), mode shapes (S) abd the modal transformation
    matrix S^-1(takes x -> r(modal coordinates) for an undamped system.

    Parameters
    ----------
    M: array
        Mass matrix
    K: array
        Stiffness matrix

    Returns
    ----------
    w: array
        The natural frequencies of the system
    P: array
        The eigenvectors of the system are.
    S: array
        The mode shapes of the system.
    Sinv: array
        The modal transformation matrix S^-1(takes x -> r(modal coordinates))

    Examples:
    >>> M = sp.array([[4, 0, 0],
    ...               [0, 4, 0],
    ...               [0, 0, 4]])
    >>> K = sp.array([[8, -4, 0],
    ...               [-4, 8, -4],
    ...               [0, -4, 4]])
    >>> w, P, S, Sinv = modes_system_undamped(M, K)
    >>> w
    array([ 0.44504187,  1.2469796 ,  1.80193774])
    """
    L = la.cholesky(M)
    Linv = la.inv(L)
    lam, P = eigen(Linv @ K @ Linv.T)
    w = sp.real(sp.sqrt(lam))
    S = Linv @ P
    Sinv = P.T @ Linv

    return w, P, S, Sinv


def modes_system(M, K, C=None):
    """
    This function will return the natural frequencies (wn), the
    damped natural frequencies (wd), the damping ratios (zeta),
    the right eigenvectors (X) and the left eigenvectors (Y) for a
    system defined by M, K and C.
    If the dampind matrix 'C' is none or if the damping is proportional,
    wd and zeta will be none and X and Y will be equal.

    Parameters
    ----------
    M: array
        Mass matrix
    K: array
        Stiffness matrix
    C: array
        Damping matrix

    Returns
    ----------
    wn: array
        The natural frequencies of the system
    wd: array
        The damped natural frequencies of the system
    zeta: array
        The damping ratios
    X: array
        The right eigenvectors
    Y: array
        The left eigenvectors

    Examples:
    >>> M = sp.array([[1, 0],
    ...               [0, 1]])
    >>> K = sp.array([[1, -0.4],
    ...               [0.4, 6]])
    >>> C = sp.array([[0.3, -4],
    ...               [4, 0.2]])
    >>> wn, wd, zeta, X, Y = modes_system(M, K, C)
    Damping is non-proportional, eigenvectors are complex.
    >>> wn
    array([ 0.5206175 ,  4.76729024,  0.5206175 ,  4.76729024])
    >>> wd
    array([ 0.50825856,  4.76531454, -0.50825856, -4.76531454])
    >>> zeta
    array([ 0.21659744,  0.02878692,  0.21659744,  0.02878692])
    >>> X
    array([[ 0.83593587+0.j        ,  0.13536177-0.00208189j,
             0.83593587-0.j        ,  0.13536177+0.00208189j],
           [ 0.00811744-0.29648107j,  0.00444279+0.15426956j,
             0.00811744+0.29648107j,  0.00444279-0.15426956j],
           [-0.09426382+0.42487157j, -0.00865558+0.6453271j ,
            -0.09426382-0.42487157j, -0.00865558-0.6453271j ],
           [ 0.14977369+0.03755828j, -0.73575270+0.j        ,
             0.14977369-0.03755828j, -0.73575270-0.j        ]])
    >>> Y
    array([[ 0.57880640-0.04653746j,  0.12020049-0.05551985j,
             0.57880640+0.04653746j,  0.12020049+0.05551985j],
           [ 0.01204879+1.25620685j, -0.06167340-0.8279964j ,
             0.01204879-1.25620685j, -0.06167340+0.8279964j ],
           [-0.00101305-0.3004544j ,  0.01513120-0.57704216j,
            -0.00101305+0.3004544j ,  0.01513120+0.57704216j],
           [ 0.10657189+0.0025583j , -0.65801243-0.00842571j,
             0.10657189-0.0025583j , -0.65801243+0.00842571j]])
    >>> C = K*2 # with proportional damping
    >>> wn, wd, zeta, X, Y = modes_system(M, K, C)
    Damping is proportional or zero, eigenvectors are real
    >>> X
    array([[-0.99677405,  0.08025891],
           [ 0.08025891, -0.99677405]])
    """

    n = len(M)

    Z = sp.zeros((n, n))
    I = sp.eye(n)
    Minv = la.inv(M)

    if (C is None or sp.all(C == 0) or # check if C has only zero entries
        la.norm(Minv @ C @ K - Minv @ K @ C, 2) < 1e-8*la.norm(Minv @ K @ C, 2)):
        w, P, S, Sinv = modes_system_undamped(M, K)
        wn = w
        wd = w
        zeta = None
        X = P
        Y = P
        print('Damping is proportional or zero, eigenvectors are real')
        return wn, wd, zeta, X, Y

    Z = sp.zeros((n, n))
    I = sp.eye(n)

    # creates the state space matrix
    A = sp.vstack([sp.hstack([Z, I]),
                   sp.hstack([-la.pinv(M) @ K, -la.pinv(M) @ C])])

    w, X = eigen(A)
    _, Y = eigen(A.T)

    wd = sp.imag(w)
    wn = sp.absolute(w)
    zeta = (-sp.real(w)/sp.absolute(w))

    Y = normalize(X, Y)

    print('Damping is non-proportional, eigenvectors are complex.')

    return wn, wd, zeta, X, Y


def response_system_undamped(M, K, x0, v0, max_time):
    """
    This function calculates the time response for an undamped system
    and returns the vector (state-space) X. The n first rows contain the
    displacement (x) and the n last rows contain velocity (v) for each
    coordinate. Each column is related to a time-step.
    The time array is also returned.

    Parameters
    ----------
    M: array
        Mass matrix
    K: array
        Stiffness matrix
    x0: array
        Array with displacement initial conditions
    v0: array
        Array with velocity initial conditions
    max_time: float
        End time

    Returns
    ----------
    t: array
        Array with the time
    X: array
        The state-space vector for each time

    Examples:
    >>> M = sp.array([[1, 0],
    ...               [0, 4]])
    >>> K = sp.array([[12, -2],
    ...               [-2, 12]])
    >>> x0 = sp.array([1, 1])
    >>> v0 = sp.array([0, 0])
    >>> max_time = 10
    >>> t, X = response_system_undamped(M, K, x0, v0, max_time)
    >>> X[:, 0] # first column of X will contain the initial conditions [x1, x2, v1, v2]
    array([ 1.,  1.,  0.,  0.])
    >>> X[:, 1] # displacement and velocities after delta t
    array([ 0.99991994,  0.99997998, -0.04001478, -0.01000397])
    """

    t = sp.linspace(0, max_time, int(250 * max_time))
    dt = t[1] - t[0]

    n = len(M)

    Z = sp.zeros((n, n))
    I = sp.eye(n, n)

    # creates the state space matrix
    A = sp.vstack([sp.hstack([Z,               I]),
                   sp.hstack([-la.pinv(M) @ K, Z])])

    # creates the x array and set the first line according to the initial
    # conditions
    X = sp.zeros((2*n, len(t)))
    X[:, 0] = sp.hstack([x0, v0])

    Ad = la.expm(A * dt)
    for i in range(len(t) - 1):
        X[:, i + 1] = Ad @ X[:, i]

    return t, X


def response_system(M, C, K, F, x0, v0, t):
    """
    This function solves the system given the initial
    displacement vector 'X0', initial velocity vector 'V0',
    the mass matrix 'M', the stiffness matrix 'M', and the damping
    matrix 'C' and force 'F'.
    T is a row vector of evenly spaced times.
    F is a matrix of forces over time, each column corresponding
    to the corresponding column of T, each row corresponding to
    the same numbered DOF.

    Parameters
    ----------
    M: array
        Mass matrix
    K: array
        Stiffness matrix
    C: array
        Damping matrix
    x0: array
        Array with displacement initial conditions
    v0: array
        Array with velocity initial conditions
    t: array
        Array withe evenly spaced times

    Returns
    ----------
    T : array
        Time values for the output.
    yout : array
        System response.
    xout : array
        Time evolution of the state vector.

    Examples:
    >>> M = sp.array([[9, 0],
    ...               [0, 1]])
    >>> K = sp.array([[27, -3],
    ...               [-3, 3]])
    >>> C = K/10
    >>> x0 = sp.array([0, 1])
    >>> v0 = sp.array([1, 0])
    >>> t = sp.linspace(0, 10, 100)
    >>> F = sp.vstack([0*t,
    ...                3*sp.cos(2*t)])
    >>> tou, yout, xout = response_system(M, C, K, F, x0, v0, t)
    >>> tou[:10]
    array([ 0.        ,  0.1010101 ,  0.2020202 ,  0.3030303 ,  0.4040404 ,
            0.50505051,  0.60606061,  0.70707071,  0.80808081,  0.90909091])
    >>> yout[:10]
    array([[ 0.        ,  1.        ,  1.        ,  0.        ],
           [ 0.1006699 ,  1.0019166 ,  0.9882704 ,  0.04164362],
           [ 0.19867119,  1.00889664,  0.94748001,  0.09763215],
           [ 0.29117815,  1.02159599,  0.87992707,  0.15230092],
           [ 0.37563287,  1.03911711,  0.78859082,  0.19072496],
           [ 0.44980625,  1.05913319,  0.67697613,  0.19968169],
           [ 0.5118431 ,  1.07810368,  0.54895551,  0.16848849],
           [ 0.56029149,  1.09156627,  0.40861572,  0.08967373],
           [ 0.59411703,  1.09448692,  0.26011522, -0.04055038],
           [ 0.61270358,  1.08164667,  0.10755703, -0.22203433]])
    >>> xout[:10]
    array([[ 0.        ,  1.        ,  1.        ,  0.        ],
           [ 0.1006699 ,  1.0019166 ,  0.9882704 ,  0.04164362],
           [ 0.19867119,  1.00889664,  0.94748001,  0.09763215],
           [ 0.29117815,  1.02159599,  0.87992707,  0.15230092],
           [ 0.37563287,  1.03911711,  0.78859082,  0.19072496],
           [ 0.44980625,  1.05913319,  0.67697613,  0.19968169],
           [ 0.5118431 ,  1.07810368,  0.54895551,  0.16848849],
           [ 0.56029149,  1.09156627,  0.40861572,  0.08967373],
           [ 0.59411703,  1.09448692,  0.26011522, -0.04055038],
           [ 0.61270358,  1.08164667,  0.10755703, -0.22203433]])
    """

    n = len(M)

    Z = sp.zeros((n, n))
    I = sp.eye(n)

    # creates the state space matrix
    A = sp.vstack([sp.hstack([Z,               I]),
                   sp.hstack([-la.pinv(M) @ K, -la.pinv(M) @ C])])
    B = sp.vstack([Z,
                   la.inv(M)])
    C = sp.eye(2*n)
    D = 0*B

    sys = signal.lti(A, B, C, D)

    IC = sp.hstack([x0, v0])
    F = F.T
    T, yout, xout = signal.lsim(sys, F, t, IC)

    return T, yout, xout


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # doctest.run_docstring_examples(frfest,globals(),optionflags=doctest.ELLIPSIS)
    # doctest.run_docstring_examples(asd,globals(),optionflags=doctest.ELLIPSIS)
    """ What this does.
    python (name of this file)  -v
    will test all of the examples in the help.

    Leaving off -v will run the tests without any output. Success will return nothing.

    See the doctest section of the python manual.
    https://docs.python.org/3.5/library/doctest.html
    """



#!/usr/bin/env python

#from distutils.core import setup
from setuptools import setup
import os
import sys

if sys.version_info < (3,5):
    sys.exit('Sorry, Python < 3.5 is not supported.')
# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name='vibration-toolbox',
      version='0.5b5',
      description='Educational code illustrating fundamentals of vibration for mechanical engineers.',
      author='Joseph C. Slater and Raphael Timbo',
      author_email='joseph.c.slater@gmail.com',
      url='https://github.com/vibrationtoolbox/vibration-toolbox',
      download_url='https://github.com/vibrationtoolbox/vibration-toolbox/archive/0.5b5.tar.gz',
      packages=['vibration-toolbox'],
      long_description = read('readme.rst'),
      keywords=['vibration','mechanical engineering'],
      install_requires=['numpy', 'scipy', 'matplotlib', 'vibration']
      )

# https://docs.python.org/3/distutils/setupscript.html#additional-meta-data



"""
The Vibration Toolbox, Python Edition

Joseph C. Slater and Raphael Timbo

__all__ = ['directory']
"""


"""
With this, `from vibration-toolbox import *` will import
all toolbox functions in the name space

`import vibration-toolbox as vt` will keep them tucked behind `vt`

`import vibration-toolbox.sdof as sdof` will tuck the sdof functions in the `sdof` name space.

"""
from .sdof import *
from .mdof import *
from .ema import *



import scipy as sp
import matplotlib.pyplot as plt


def frf(x, f, dt):
    """
    This function will return the frequency response
    function (H(iw)) of the sampled data.

    Parameters
    ----------
    x: array
        Array with the displacement data
    f: array
        Array with the force data
    dt: float
        Time step of the sampled data
    n: int
        Number of points in the fft

    Returns
    ----------
    freq: array
        Driving frequencies
    mag: array
        Magnitude of the frequency response function
    ang: array
        Phase of the frequency response function
    coh: array
        Coherence function

        Plot with the frf magnitude, phase and
        coherence.
    Examples:
    >>> # First we need to load the sampled data which in a .mat file
    >>> import vibration-toolbox as vt
    >>> import scipy.io as sio
    >>> import os
    >>> #print(os.getcwd())
    >>> #print(os.listdir('vibration-toolbox/data'))
    >>> data = sio.loadmat('vibration-toolbox/data/frf_data1.mat')
    >>> #print(data)
    >>> # Data is imported as arrays. We need to modify then to fit our function
    >>> x = data['x']
    >>> x = x.reshape(len(x))
    >>> f = data['f']
    >>> f = f.reshape(len(f))
    >>> dt = data['dt']
    >>> dt = float(dt)
    >>> # Now we are able to call the function
    >>> freq, mag, ang, coh = vt.frf(x, f, dt)
    >>> mag[10]   # doctest: +ELLIPSIS
    1.018394853080...
    """

    w = sp.sin(sp.pi*sp.arange(len(f))/len(f))**2 # window
    # apply window
    xw = x*w
    fw = f*w
    # take ffts
    FX = sp.fft(xw)
    FF = sp.fft(fw)
    # calculate the spectral densities
    SXF = FF*sp.conj(FX)
    SXX = FX*sp.conj(FX)
    SFF = FF*sp.conj(FF)
    SFX = FX*sp.conj(FF)
    # calculate the transfer functions
    TXF = SXX/SXF
    TXF2 = SFX/SFF

    lt = len(TXF)//2
    freq = sp.arange(lt)/(2*lt*dt)

    TXF = TXF[:lt]
    mag = sp.absolute(TXF)
    ang = sp.angle(TXF)*180/sp.pi

    coh = (sp.absolute(SXF)**2)/(SXX*SFF)
    coh = sp.real(coh)

    # plot H(w)
    fig = plt.figure(figsize=(8,6))
    ax1 = fig.add_subplot(311)
    ax2 = fig.add_subplot(312, sharex=ax1)
    ax3 = fig.add_subplot(313, sharex=ax2)
    fig.tight_layout()

    ax1.set_title('$H(\omega)$ - Magnitude')
    ax2.set_title('$H(\omega)$ - Phase')
    ax3.set_title('$H(\omega)$ - Coherence')
    ax3.set_xlabel('Frequency (Hz)')
    ax3.set_ylim(0,2)

    ax1.semilogy(freq, mag)
    ax2.plot(freq, ang)
    ax3.plot(freq, coh[:lt])

    _ = plt.show()

    return freq, mag, ang, coh


if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)

    """ What this does.
    python (name of this file)  -v
    will test all of the examples in the help.

    Leaving off -v will run the tests without any output. Success will return nothing.

    See the doctest section of the python manual.
    https://docs.python.org/3.5/library/doctest.html
    """
